/*
	r_light.c

	(description)

	Copyright (C) 1996-1997  Id Software, Inc.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place - Suite 330
		Boston, MA  02111-1307, USA

	$Id$
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "QF/render.h"

#include "r_local.h"


void
R_AnimateLight (void)
{
	int         i, j, k;

	// light animations
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	i = (int) (r_realtime * 10);
	for (j = 0; j < MAX_LIGHTSTYLES; j++) {
		if (!r_lightstyle[j].length) {
			d_lightstylevalue[j] = 256;
			continue;
		}
		k = i % r_lightstyle[j].length;
		k = r_lightstyle[j].map[k] - 'a';
		k = k * 22;
		d_lightstylevalue[j] = k;
	}
}

/*
	DYNAMIC LIGHTS
*/

void
R_RecursiveMarkLights (vec3_t lightorigin, dlight_t *light, int bit,
					   mnode_t *node)
{
	mplane_t   *splitplane;
	float       ndist, maxdist;
	msurface_t *surf;
	int         i;

	maxdist = light->radius * light->radius;
loc0:
	if (node->contents < 0)
		return;

	splitplane = node->plane;
	ndist = DotProduct (lightorigin, splitplane->normal) - splitplane->dist;

	if (ndist > light->radius) {
		if (node->children[0]->contents >= 0) {
			node = node->children[0];
			goto loc0;
		}
		return;
	}
	if (ndist < -light->radius) {
		if (node->children[1]->contents >= 0) {
			node = node->children[1];
			goto loc0;
		}
		return;
	}
	// mark the polygons
	surf = r_worldentity.model->surfaces + node->firstsurface;
	for (i = 0; i < node->numsurfaces; i++, surf++) {
		if (surf->dlightframe != r_framecount) {
			surf->dlightbits = 0;
			surf->dlightframe = r_framecount;
		}
		surf->dlightbits |= bit;
	}

	if (node->children[0]->contents >= 0) {
		if (node->children[1]->contents >= 0)
			R_RecursiveMarkLights (lightorigin, light, bit, node->children[1]);
		node = node->children[0];
		goto loc0;
	} else if (node->children[1]->contents >= 0) {
		node = node->children[1];
		goto loc0;
	}
}

static void
mark_surfaces (msurface_t *surf, vec3_t lightorigin,  dlight_t *light,
			   int bit)
{
	float      dist;
#if 1
	float      dist2, d;
	float      maxdist = light->radius * light->radius;
	vec3_t     impact;
#endif

	dist = PlaneDiff(lightorigin, surf->plane);
	if (surf->flags & SURF_PLANEBACK)
		dist = -dist;
	if ((dist < -0.25f && !(surf->flags & SURF_LIGHTBOTHSIDES))
		|| dist > light->radius)
		return;
#if 1
	dist2 = dist * dist;
	dist = -dist;
	VectorMA (light->origin, dist, surf->plane->normal, impact);

	d = DotProduct (impact, surf->texinfo->vecs[0])
		+ surf->texinfo->vecs[0][3] - surf->texturemins[0];
	if (d < 0) {
		dist2 += d * d;
		if (dist2 >= maxdist)
			return;
	} else {
		d -= surf->extents[0] + 16;
		if (d > 0) {
			dist2 += d * d;
			if (dist2 >= maxdist)
				return;
		}
	}
	d = DotProduct (impact, surf->texinfo->vecs[1])
		+ surf->texinfo->vecs[1][3] - surf->texturemins[1];
	if (d < 0) {
		dist2 += d * d;
		if (dist2 >= maxdist)
			return;
	} else {
		d -= surf->extents[1] + 16;
		if (d > 0) {
			dist2 += d * d;
			if (dist2 >= maxdist)
				return;
		}
	}
#endif
	if (surf->dlightframe != r_framecount) {
		surf->dlightbits = 0;
		surf->dlightframe = r_framecount;
	}
	surf->dlightbits |= bit;
}

void
R_MarkLights (vec3_t lightorigin, dlight_t *light, int bit, model_t *model)
{
	mleaf_t    *pvsleaf = Mod_PointInLeaf (lightorigin, model);

	if (!pvsleaf->compressed_vis) {
		mnode_t *node = model->nodes + model->hulls[0].firstclipnode;
		R_RecursiveMarkLights (lightorigin, light, bit, node);
	} else {
		float       radius = light->radius;
		vec3_t      mins, maxs;
		int         leafnum = 0;
		byte       *in = pvsleaf->compressed_vis;
		byte        vis_bits;

		mins[0] = lightorigin[0] - radius;
		mins[1] = lightorigin[1] - radius;
		mins[2] = lightorigin[2] - radius;
		maxs[0] = lightorigin[0] + radius;
		maxs[1] = lightorigin[1] + radius;
		maxs[2] = lightorigin[2] + radius;
		while (leafnum < model->numleafs) {
			int         i;
			if (!(vis_bits = *in++)) {
				leafnum += (*in++) * 8;
				continue;
			}
			for (i = 0; i < 8 && leafnum < model->numleafs; i++, leafnum++) {
				int      m;
				mleaf_t *leaf  = &model->leafs[leafnum + 1];
				if (!(vis_bits & (1 << i)))
					continue;
				if (leaf->visframe != r_visframecount)
					continue;
				if (leaf->mins[0] > maxs[0] || leaf->maxs[0] < mins[0]
					|| leaf->mins[1] > maxs[1] || leaf->maxs[1] < mins[1]
					|| leaf->mins[2] > maxs[2] || leaf->maxs[2] < mins[2])
					continue;
				if (leaf->dlightframe != r_framecount) {
					leaf->dlightbits = 0;
					leaf->dlightframe = r_framecount;
				}
				leaf->dlightbits |= bit;
				for (m = 0; m < leaf->nummarksurfaces; m++) {
					msurface_t *surf = leaf->firstmarksurface[m];
					if (surf->visframe != r_visframecount)
						continue;
					mark_surfaces (surf, lightorigin, light, bit);
				}
			}
		}
	}
}


void
R_PushDlights (vec3_t entorigin)
{
	int         i;
	dlight_t   *l;
	vec3_t      lightorigin;

	l = r_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, l++) {
		if (l->die < r_realtime || !l->radius)
			continue;
		VectorSubtract (l->origin, entorigin, lightorigin);
		R_MarkLights (lightorigin, l, 1 << i, r_worldentity.model);
	}
}


/*
	LIGHT SAMPLING
*/


int
RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
	int         r;
	float       front, back, frac;
	int         side;
	mplane_t   *plane;
	vec3_t      mid;
	msurface_t *surf;
	int         s, t, ds, dt;
	int         i;
	mtexinfo_t *tex;
	byte       *lightmap;
	unsigned int scale;
	int         maps;

	if (node->contents < 0)
		return -1;						// didn't hit anything

	// calculate mid point
// FIXME: optimize for axial
	plane = node->plane;
	front = DotProduct (start, plane->normal) - plane->dist;
	back = DotProduct (end, plane->normal) - plane->dist;
	side = front < 0;

	if ((back < 0) == side)
		return RecursiveLightPoint (node->children[side], start, end);

	frac = front / (front - back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side   
	r = RecursiveLightPoint (node->children[side], start, mid);
	if (r >= 0)
		return r;						// hit something

	if ((back < 0) == side)
		return -1;						// didn't hit anything

	// check for impact on this node
	surf = r_worldentity.model->surfaces + node->firstsurface;
	for (i = 0; i < node->numsurfaces; i++, surf++) {
		if (surf->flags & SURF_DRAWTILED)
			continue;					// no lightmaps

		tex = surf->texinfo;

		s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
		t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

		if (s < surf->texturemins[0] || t < surf->texturemins[1])
			continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if (ds > surf->extents[0] || dt > surf->extents[1])
			continue;

		if (!surf->samples)
			return 0;

		ds >>= 4;
		dt >>= 4;

		lightmap = surf->samples;
		r = 0;
		if (lightmap) {

			lightmap += dt * ((surf->extents[0] >> 4) + 1) + ds;

			for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255;
				 maps++) {
				scale = d_lightstylevalue[surf->styles[maps]];
				r += *lightmap * scale;
				lightmap += ((surf->extents[0] >> 4) + 1) *
					((surf->extents[1] >> 4) + 1);
			}

			r >>= 8;
		}

		return r;
	}

	// go down back side
	return RecursiveLightPoint (node->children[!side], mid, end);
}


int
R_LightPoint (vec3_t p)
{
	vec3_t      end;
	int         r;

	if (!r_worldentity.model->lightdata)
		return 255;

	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 2048;

	r = RecursiveLightPoint (r_worldentity.model->nodes, p, end);

	if (r == -1)
		r = 0;

	if (r < r_refdef.ambientlight)
		r = r_refdef.ambientlight;

	return r;
}
